package org.nlogo.extensions.netcdf;

/* Standard NetLogo imports */
import org.nlogo.api.CompilerException;
import org.nlogo.api.LogoException;
import org.nlogo.api.ExtensionException;
import org.nlogo.api.Argument;
import org.nlogo.api.Syntax;
import org.nlogo.api.Dump;
import org.nlogo.api.Context;
import org.nlogo.api.LogoList;
import org.nlogo.api.LogoListBuilder;
import org.nlogo.api.DefaultReporter;
import org.nlogo.api.DefaultCommand;

/* Imports for working with matrix extension */
// // Does not work as LogoMatrix is private
// import org.nlogo.extensions.matrix.MatrixExtension.LogoMatrix;

/* import netcdf */
import ucar.ma2.Array;
import ucar.ma2.IndexIterator;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.NetcdfFile;
import ucar.nc2.dataset.NetcdfDataset;

/* Standard java stuff */
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class NetCDFExtension
    extends org.nlogo.api.DefaultClassManager {

    private final NetcdfFileManager netcdfFM = new NetcdfFileManager();
    
    // This will be the class to manage open netcdf files and do the
    // direct interaction with java netcdf library
    private final strictfp class NetcdfFileManager
    // NetcdfFile is sufficiently different from java.io.File that it seems
    // difficult to actually implement FileManager here, but perhaps
    // someone with more experience will say otherwise.
    // implements org.nlogo.nvm.FileManager 
    {

	private final List<NetcdfFile> openFiles =
	    new ArrayList<NetcdfFile>();
	private NetcdfFile currentFile;
	
	public int numberOfFiles() {
	    return openFiles.size();
	}

	private String currentFile() {
	    return currentFile.getLocation();
	}

	public boolean hasCurrentFile() {
	    return (currentFile != null);
	}

	private boolean isFileOpen(String location) {
	    return (findOpenFile(location) != null);
	}

	public NetcdfFile findOpenFile(String location) {
	    // Note this will only match if location is precisely the same
	    // i.e., location="myfile.nc" is not the same as "./myfile.nc"
	    // even though these may indeed refer to the same file.
	    // Ideally, this should be fixed, but seems tedious.
	    for (NetcdfFile nextFile : openFiles) {
		if (location.equals(nextFile.getLocation())) {
		    return nextFile;
		}
	    }

	    return null;
	}

	private boolean isFileCurrent(NetcdfFile checkFile) {
	    return (currentFile == checkFile);
	}

	private void setCurrentFile(NetcdfFile newFile) {
	    currentFile = newFile;
	}

	// public boolean fileExists(String location) {
	//     return new NetcdfFile.canOpen(location);
	// }

	public void openFile(String location)
	    throws IOException {

	    if (location == null) {
		closeAllFiles(); // Call this here before exception
		throw new IOException("Location is null!");
	    }

	    // Try to find it among already open files
	    NetcdfFile newFile = findOpenFile(location);

	    if (newFile == null) {
		newFile = NetcdfFile.open(location);
		openFiles.add(newFile);
	    }

	    setCurrentFile(newFile);
	}

	public void closeCurrentFile()
	    throws IOException {
	    if (!hasCurrentFile()) {
		closeAllFiles(); // Call this here before exception
		throw new IOException("There is no file to close");
	    }
	    closeFile(currentFile);
	    setCurrentFile(null);
	}

	private void closeFile(NetcdfFile closeFile)
	    throws IOException {
	    closeFile.close();
	    openFiles.remove(closeFile);
	}

	public Array readVariable(String varname)
	    throws IOException {
	    return currentFile.findVariable(varname).read();
	}

	public Array readVariable(String varname, int[] origin, int[] shape)
	    throws IOException, InvalidRangeException {
	    return currentFile.findVariable(varname).read(origin,shape);
	}

	public void closeAllFiles() 
	    throws IOException {
	    StringBuilder errors = new StringBuilder();

	    Iterator<NetcdfFile> files = openFiles.iterator();
	    while (files.hasNext()) {
		NetcdfFile f = files.next();
		try {
		    closeFile(f);
		} catch (IOException e) {
		    errors.append("Unable to close netcdf " + 
				  f.getLocation() + "\n" +
				  e.getMessage() + "\n" );
		}
		files = openFiles.iterator();
	    }
	    setCurrentFile(null);

	    if (errors.length() > 0) {
		throw new IOException(errors.toString());
	    }
	}

	// Returns a list of each file location with current one last
	public List<String> filesList() {
	    List<String> locations = new ArrayList<String>();

	    for (NetcdfFile nextFile : openFiles) {
		if (!isFileCurrent(nextFile)) {
		    locations.add(nextFile.getLocation());
		}
	    }
	    // Stick current netcdf at end for when system reloads
	    if (hasCurrentFile()) {
		locations.add(currentFile.getLocation());
	    }
	    return locations;   
	}
	
	public void openFilesList(List<String> locations) 
	    throws IOException {
	    StringBuilder errors = new StringBuilder();

	    for (String location : locations) {
		try {
		    openFile(location);
		} catch (IOException e) {
		    errors.append("Error opening netcdf " + location + 
				  ". This netcdf will be ignored.\n" +
				  e.getMessage() + "\n");
		}
		
	    }
	    
	    if (errors.length() > 0) {
		throw new IOException(errors.toString());
	    }
	}
    }
    
    /// For exporting and importing world

    public StringBuilder exportWorld() {
	StringBuilder buffer = new StringBuilder();
	List<String> locations = netcdfFM.filesList();

	for (String location : locations) {
	    buffer.append( location + "\n" );
	}
	return buffer;
    }
    
    public void importWorld(java.util.List<String[]> lines, org.nlogo.api.ExtensionManager reader,
			    org.nlogo.api.ImportErrorHandler handler)
	throws ExtensionException {
	// First be sure to close all already open files if any
	try {
	    netcdfFM.closeAllFiles();
	} catch (IOException e) {
	    handler.showError("Error in importWorld of netcdf extension.",
			      e.getMessage(),"Continuing bravely.");
	}

	// Now convert String[] list to String list
	List<String> locations = new ArrayList<String>();
	for (String[] line : lines) {
	    locations.add(line[0]);
	}

	// Open files, showing any resulting error messages
	try {
	    netcdfFM.openFilesList(locations);
	} catch (IOException e) {
	    handler.showError("Error in importWorld of netcdf extension.",
			      e.getMessage(),"Continuing bravely.");
	}
    }
    
    /// Loading and unloading of extension
    
    public void load(org.nlogo.api.PrimitiveManager primManager) {
	// netcdf:open StringNetcdfLocation
	primManager.addPrimitive("open", new openNetcdf());

	// netcdf:number-open => integer number of open files
	primManager.addPrimitive("number-open", new numOpenNetcdfs());

	// netcdf:list-of-open => LogoList of String netcdf locations
	primManager.addPrimitive("list-of-open", new listOpenNetcdfs());

	// netcdf:current => String netcdf location of active netcdf
	primManager.addPrimitive("current", new currentNetcdf());

	// netcdf:variable-to-list StringNetcdfLocation OriginList ExtentList =>
	// nested LogoList of data
	primManager.addPrimitive("variable-to-list", new readNetcdfToList());

	// netcdf:entire-variable-to-list StringNetcdfLocation =>
	// nested LogoList of data
	primManager.addPrimitive("entire-variable-to-list", new readNetcdfToListFull());

	// netcdf:close - closes active netcdf
	primManager.addPrimitive("close", new closeNetcdf());
    }
    
    public void unload(org.nlogo.api.ExtensionManager em) 
	throws ExtensionException { 
	try {
	    netcdfFM.closeAllFiles();
	} catch (Exception e) {
	    System.err.println("Error closing all netcdf files: " + 
			       e.getMessage());
	}
    }

    /// Primitive definitions below
    
    
    public class numOpenNetcdfs extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax
		(new int[]{},Syntax.NumberType());
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public Object report(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    return Double.valueOf(netcdfFM.numberOfFiles());
	}
    }
    
    public class listOpenNetcdfs extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax
		(new int[]{},Syntax.ListType());
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public Object report(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    return LogoList.fromJava(netcdfFM.filesList());
	}
    }
    
    public class currentNetcdf extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax
		(new int[]{},Syntax.StringType());
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public Object report(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    if (netcdfFM.hasCurrentFile()) {
		return netcdfFM.currentFile();
	    }
	    return new String();
	}
    }
    
    public class openNetcdf extends DefaultCommand {
	public Syntax getSyntax() {
	    return Syntax.commandSyntax
		(new int[]{Syntax.StringType()});
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public void perform(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    String location = args[0].getString();
	    try {
		netcdfFM.openFile(location);
	    } catch (IOException e) {
		throw new ExtensionException("Could not open file " + location);
	    }
	}
    }

    public class closeNetcdf extends DefaultCommand {
	public Syntax getSyntax() {
	    return Syntax.commandSyntax
		(new int[]{});
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public void perform(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    if (!netcdfFM.hasCurrentFile()) {
		throw new ExtensionException
		    ("No netcdf open and active.  Perhaps reopen to activate");
	    }

	    try {
		netcdfFM.closeCurrentFile();
	    } catch (IOException e) {
		throw new ExtensionException("Could not close current netcdf file");
	    }
	}
    }

    public class readNetcdfToList extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax
		(new int[]{Syntax.StringType(),Syntax.ListType(),Syntax.ListType()},
		 Syntax.ListType());
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public Object report(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    String vn = args[0].getString();

	    LogoList originL = args[1].getList();
	    int[] origin = new int[originL.size()];  
	    int i=0;
	    for (Object n : originL) {
		origin[i] = (int) ((Number) n).doubleValue();
		i++;
	    }

	    LogoList shapeL = args[2].getList();
	    int[] shape = new int[shapeL.size()];  
	    i=0;
	    for (Object n : shapeL) {
		shape[i] = (int) ((Number) n).doubleValue();
		i++;
	    }

	    try {
		return ArrayToNestedList(netcdfFM.readVariable(vn,origin,shape).reduce());
	    } catch (Exception e) {
		throw new ExtensionException("Could not read from current netcdf file or convert to list");
	    }
	}
    }

    public class readNetcdfToListFull extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax
		(new int[]{Syntax.StringType()},
		 Syntax.ListType());
	}
	
	public String getAgentClassString() {
	    return "OTPL";
	}
	
	public Object report(Argument args[], Context context)
	    throws ExtensionException, LogoException {
	    String vn = args[0].getString();

	    try {
		return ArrayToNestedList(netcdfFM.readVariable(vn).reduce());
	    } catch (IOException e) {
		throw new ExtensionException("Could not read from current netcdf file or convert to list");
	    }
	}
    }

    public LogoList ArrayToNestedList(Array array) {
	int nd = array.getRank();

	LogoListBuilder lst = new LogoListBuilder();
	if (nd>1) {
	    int[] shape = array.getShape();
	    for (int k=0; k < shape[0]; k++) {
		LogoList ll = ArrayToNestedList(array.slice(0,k));
		lst.add(ll);
	    }
	} else {
	    IndexIterator i = array.getIndexIterator();		
	    while ( i.hasNext() ) {
		Object v = i.next();
		if (v instanceof Boolean) {
		    lst.add(v);
		} else {
		    lst.add(((Number) v).doubleValue());
		}
	    }
	}
	return lst.toLogoList();
    }

    // // Not working because LogoMatrix is private and I still don't
    // // quite understand how copyToXDJavaArray works.

    // public LogoMatrix ArrayToLogoMatrix(Array array) {
    // 	int nd = array.getRank();
    // 	int[] shape = array.getShape();
    // 	long size = array.getSize();
    // 	long numRows;
    // 	long numCols;

    // 	if (nd == 2) {
    // 	    numRows = shape[0];
    // 	    numCols = shape[1];
    // 	} else {
    // 	    numRows = 1;
    // 	    numCols = size;
    // 	}

    // 	double[][] dArray = new double[numRows][numCols];	

    // 	if (nd == 2) {
    // 	    array.copyToNDJavaArray(dArray);
    // 	} else {
    // 	    array.copyTo1DJavaArray(dArray[0]);
    // 	}

    // 	return new LogoMatrix(new Jama.Matrix(dArray));
    // }
    
}
